// IvCameraRep.h++

// In this object, "reference transform" and "coordinate transform" are
// of the camera frame relative to the world.  Thusly, rt is where the camera
// is relative to the world, ct is where the world is relative to the camera

#ifndef _IvCameraRep_H_
#define _IvCameraRep_H_

// #include <Inventor/draggers/SoJackDragger.h>
#include <Inventor/draggers/SoTransformBoxDragger.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/SbString.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTransformSeparator.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/actions/SoGetMatrixAction.h>

class IvCameraRep
{
    // Utility 
    static int initClass();
    static const SbName rootName;
    static const SbName copName;

    // Geometry for image plane and COP
    static float axisCoordinateValues[20][3];
    // static const long axisLineCoords[33];
    static const int32_t axisLineCoords[33];
    static SoSeparator* axisGeometry;
    static SoCoordinate3* axisCoordinates;
    static SoIndexedLineSet* axisLines;
    static SoSphere* copGeometry;

    // For driving my geometry from my manipulators
    static SoDraggerCB copDragValueCB;
    static SoDraggerCB imagePlaneDragValueCB;

    // For driving other nodes from my manipulators
    static SoDraggerCB interiorOrientationDraggerCB;
    static SoDraggerCB referenceTransformDraggerCB;
    static SoDraggerCB coordinateTransformDraggerCB;

    // For driving my geometry from other nodes
    static SoSensorCB interiorOrientationSensorCB;
    static SoSensorCB referenceTransformSensorCB;
    static SoSensorCB coordinateTransformSensorCB;
    SoNodeSensor* interiorOrientationSensor;
    SoNodeSensor* relativeOrientationSensor;

    // Scene graph
    // Graphics OF camera geometry, for display in 3-D viewer
    // SEP
    //  |
    // IP-DRAG...MAT..REL-OR..AXIS-GEOM..COP..............COP-DRAG
    //  |                                 |                  |
    // DRAGXF...DRAGGER            INT-OR...COP-GEOM  DRAGXF...DRAGGER
    SoSeparator* root;
    SoSeparator* imagePlaneDraggerRoot;
    SoTransform* imagePlaneDraggerTransform;
    // SoJackDragger* imagePlaneDragger;
    SoTransformBoxDragger* imagePlaneDragger;
    SoMaterial* material;
    SoTransform* referenceTransform;
    // axis geometry is static
    SoTransformSeparator* copRoot;
    SoTransform* interiorOrientation;
    // cop geometry is static
    SoSeparator* copDraggerRoot;
    SoTransform* copDraggerTransform;
    SoTranslate1Dragger* copDragger;
    SoDraggerCB* copDragUserCB;
    void* copDragUserData;
    SoDraggerCB* imagePlaneDragUserCB;
    void* imagePlaneDragUserData;

    int rayCount;
    SoSeparator* rayGeometry;
    SoCoordinate3* rayCoordinates;
    SoIndexedLineSet* rayLines;

public:
    IvCameraRep::IvCameraRep();
    IvCameraRep::IvCameraRep( const IvCameraRep& );
    IvCameraRep::~IvCameraRep();
    SoSeparator* getNode() { return root; }
    inline void setColor( float r, float g, float b );

    inline void setInteriorOrientation( float beta );
    inline float setInteriorOrientation( SoPerspectiveCamera* );
    inline float getFocalLength();

    inline void setReferenceTransform( SbMatrix& );
    inline void addInteriorOrientationCB( SoDraggerCB*, void* );
    inline void addRelativeOrientationCB( SoDraggerCB*, void* );
    
    void setReferenceTransform( float tx, float ty, float tz,
				float q0, float q1, float q2, float q3 );
    void setCoordinateTransform( float tx, float ty, float tz,
				 float q0, float q1, float q2, float q3 );
    inline SoTransform* getReferenceTransform(){ return referenceTransform; }

    int select( SoPath* );
    int deselect( SoPath* );

    void attachInteriorOrientation( SoPerspectiveCamera* );
    void attachReferenceTransform( SoTransform* );
    void attachCoordinateTransform( SoTransform* );
    void driveInteriorOrientation( SoPerspectiveCamera* );
    void driveReferenceTransform( SoTransform* );
    void driveCoordinateTransform( SoTransform* );

    void setSize( int h, int w );

    inline int drawRay( int i, float u, float v, float depth = 4 );
};

inline
void
IvCameraRep::setColor( float r, float g, float b )
{
    material->ambientColor.setValue( r, g, b );
    material->diffuseColor.setValue( r, g, b );
}

inline
void
IvCameraRep::setInteriorOrientation( float beta )
{
    if( beta < .01 )
       beta = .01;
    interiorOrientation -> translation.setValue( 0., 0., -1./beta );
    rayCoordinates -> point.set1Value( 0, 0., 0., -1/beta );
}

inline
void
IvCameraRep::setReferenceTransform( SbMatrix& m )
{
    referenceTransform -> setMatrix( m );
}

inline
float
IvCameraRep::getFocalLength()
{
    return( interiorOrientation -> translation.getValue()[2] * -1. );
}

inline
void
IvCameraRep::addInteriorOrientationCB( SoDraggerCB* fn, void* data )
{
    copDragUserCB = fn;
    copDragUserData = data;
}

inline
void
IvCameraRep::addRelativeOrientationCB( SoDraggerCB* fn, void* data )
{
    imagePlaneDragUserCB = fn;
    imagePlaneDragUserData = data;
}


float
IvCameraRep::setInteriorOrientation( SoPerspectiveCamera* pc )
{
    float beta = tan( pc->heightAngle.getValue() / 2. ) / .375;
    setInteriorOrientation( beta );
    return beta;
}

inline
int
IvCameraRep::drawRay( int i, float u, float v, float depth )
{
    float f = getFocalLength();
    float s = ( 1. + depth/f );
    if( i < 0 || i > rayCount )
       return 1;
    if( i == rayCount )
       rayCoordinates -> point.setNum( ++rayCount + 1 );
    rayCoordinates -> point.set1Value( i+1, s*u, s*v, depth );
    int32_t conn[3];
    conn[0] = 0;
    conn[1] = i+1;
    conn[2] = -1;
    rayLines -> coordIndex.setValues( i*3, 3, conn );
    return 0;
}

#endif /* _IvCameraRep_H_ */
