// qmEigenStructure.c++

#include "QuickMath/qmEigenStructure.h++"

qmEigenStructure::qmEigenStructure( const qmMatrixRef& m )
  : eigenVectors( m.getWidth(), m.getWidth() ),
    eigenValues( m.getWidth() )
{
    set( m );
}

qmEigenStructure&
qmEigenStructure::set( const qmMatrixRef& m )
{
#if QmDoBoundaryChecking
    QmAssert( m.getHeight() == m.getWidth() &&
	    m.getWidth() == getWidth() );
#endif
    eigenVectors.set( m.getWidth(), m.getWidth() );
    eigenValues.set( m.getWidth() );
    eigenSymmetric( m, eigenValues, eigenVectors );
    return *this;
}

static
inline
int
eigenSymmetric2x2f( real *M, real* L, real v[2][2] )
{
    const double minAcceptableNorm = 1e-15;

    real& a = M[0];
    real& b = M[1];
    real& c = M[2];
    real& d = M[3];

    // Compute eigenvalues, solve quadratic equation
    // ------------------------------------------------
    double aplusd = a + d;
    double alessd = a - d;
    double rad = sqrt( alessd*alessd + 4*b*b );
    L[0] = .5 * ( aplusd - rad );
    L[1] = .5 * ( aplusd + rad );

    // Compute eigenvectors
    // ------------------------------------------------
    //  [ a-l   b  ][ x1 ]
    //  [  b   d-l ][ x2 ] is zero
    // 1. find the most well-conditioned row (2 evalues, 2 rows each)
    double norms[2][2];
    double alessl[2];
    double dlessl[2];
    alessl[0] = a - L[0];
    dlessl[0] = d - L[0];
    alessl[1] = a - L[1];
    dlessl[1] = d - L[1];
    norms[0][0] = alessl[0]*alessl[0] + b*b;
    norms[0][1] = b*b + dlessl[0]*dlessl[0];
    norms[1][0] = alessl[1]*alessl[1] + b*b;
    norms[1][1] = b*b + dlessl[1]*dlessl[1];
    double max_norm = norms[0][0];
    int whicheig = 0;
    int whichrow = 0;
    for( int i = 0; i < 2; i++ )
       for( int j = 0; j < 2; j++ )
	  if( norms[i][j] > max_norm )
	     {
		 max_norm = norms[i][j];
		 whicheig = i;
		 whichrow = j;
	     }
    // 2. now check that this norm is good enough precision...
    max_norm = sqrt( max_norm );
    if( max_norm < minAcceptableNorm )
       { // ...if not, we have a circular distribution
	   v[0][0] = v[1][1] = 1.;
	   v[0][1] = v[1][0] = 0.;
	   return 1;
       }
    // ...if so, ...
    // 3. normalize the best row and find the eigenvectors
    if( 0 == whichrow )
       {
	   v[whicheig][0] = -b / max_norm;
	   v[whicheig][1] = alessl[whicheig] / max_norm;
       }
    else
       {
	   v[whicheig][0] = dlessl[whicheig] / max_norm;
	   v[whicheig][1] = -b / max_norm;
       }
    v[1-whicheig][0] = v[whicheig][1];
    v[1-whicheig][1] = -v[whicheig][0];
    if( b != c )
       return -1; // not a symmetric matrix
    return 0;
}

int
qmEigenStructure::eigenSymmetric2( const qmMatrixRef& sym2x2Mat,
				   qmVectorRef& eValOut, 
				   qmMatrixRef& eVecOut )
{
#ifdef QmDoBoundaryChecking
    QmAssert( sym2x2Mat.getElementCount() == 4 && 
	    sym2x2Mat.getWidth() == 2 &&
	    sym2x2Mat( 0, 1 ) == sym2x2Mat( 1, 0 ) );
#endif
    real v[2][2];
    int result = eigenSymmetric2x2f( sym2x2Mat._getArray(), eValOut, v );
    eVecOut( 0, 0 ) = v[0][0];
    eVecOut( 0, 1 ) = v[0][1];
    eVecOut( 1, 0 ) = v[1][0];
    eVecOut( 1, 1 ) = v[1][1];
    return result;
}

// jacobi.c++
// modified jacobi.c for C++ by Ali Azarbayejani
/* 
 copyright 1992, M.I.T.  All Rights Reserved.

*/
/*
  File: jacobi.c
  Author: Matthew Turk
  Date: 3/22/91
  Modified:  Thad Starner 10/6/91
  For further information see:
  
  Turk, M., and Pentland, A., (1991) Eigenfaces for Recognition, 
  Journal of Cognitive Neuroscience, Vol. 3, No. 1, pp. 71-86.
  
*/
/*
   Use Jacobi Transformations of a symetric matrix to obtain the 
   normalized eigenvectors and eigen values 
   -- Taken from NUMERICAL RECIPES IN C (pp 364-366) --
*/

#define EPS 0.000000001
#define ROTATE(a, i, j, k, l, n) g=a[i*n+j]; h=a[k*n+l]; a[i*n+j] = g-s*(h+g*tau); \
  a[k*n+l] = h+s*(g-h*tau);
      
static
void
jacobi( real* a, int n, real* d, real* v, int* nrot )

/* 
  Computes all eigenvalues and eigenvectors of a real symmetric matrix A
  On output, elements of A above the diagonal are destroyed,
  D returns the eigenvalues of A, and V is a matrix whose columns are the
  eigenvectors of A.
*/

  {
  
  int j;
  int iq;
  int ip;
  int ip_times_n;
  int i;
  real tresh;
  real theta;
  real tau;
  real t;
  real sm;
  real s;
  real h;
  real g;
  real c;
  real* b;
  real* z;
  real* vector();
  
  b = new real[n];
  z = new real[n];
  
  for(ip_times_n=0, ip=0; ip<n; ++ip, ip_times_n+=n)  
      {
      
      /* Initialize the identity matrix */
      for(iq=0; iq<n; ++iq)v[ip_times_n + iq] = 0.0 ;
      v[ip_times_n + ip] = 1.0 ;
      
      /* Initailize b and d to be diagonal of a */
      b[ip] = d[ip] = a[ip_times_n + ip];
      z[ip] = 0.0 ;
      }
      
  *nrot = 0 ;
  for(i=0;i<50;++i)
      {
      /* Sum off-diagonal elements */
      sm=0.0 ;

      for(ip_times_n=0,ip=0;ip<n-1;ip++,ip_times_n+=n)
	for(iq=ip+1;iq<n;iq++)
	  sm += fabs(a[ip_times_n + iq]);
	  
      /*  If we have converged,  free the working vectors and return. */
      if(sm == 0.0)
	  {
	      delete [] b;
	      delete [] z;
	      return;
	  }
      
      /* tresh is different on first three iterations...*/
      tresh=(i<3) ? 0.2*sm/(n*n) : 0.0 ;
      
      for(ip_times_n=0,ip=0;ip<n-1;ip++,ip_times_n+=n)
	  {
	  for(iq=ip+1;iq<n;++iq)
	      {
	      g=100.0*fabs(a[ip_times_n + iq]);

	      /* After four sweeps, skip the rotation if the off-diagonal element is small */
	      /* This test is taken directly from the text and looks a little suspect to me... */

	      if(i > 3 && g < EPS)
		a[ip_times_n + iq] = 0.0 ;

	      else if(fabs(a[ip_times_n+iq]) > tresh) 
		  {
		  h=d[iq]-d[ip];
		  if(g < EPS)
		    t = (fabs(a[ip_times_n+iq]) > EPS) ? (a[ip_times_n+iq])/h : 0.0 ; 
		  else
		      { 
		      theta=(fabs(h) < EPS) ? 0.0 : 0.5*h/(a[ip_times_n+iq]);
		      t=1.0/(fabs(theta)+sqrt(1.0+theta*theta));
		      if(theta < 0.0)
			t = -t ; 
		      } 
		  c=1.0/sqrt(1.0+t*t);
		  s=t*c;
		  tau=s/(1.0+c);
		  
		  h=t*a[ip_times_n+iq];
		  z[ip] -= h;
		  z[iq] += h;
		  d[ip] -= h;
		  d[iq] += h;
		  a[ip_times_n+iq]=0.0;
		  
		  for(j=0;j<ip;j++)
		      {
		      ROTATE(a,j,ip,j,iq,n);
		      }
		  for(j=ip+1;j<iq;j++)
		      {
		      ROTATE(a,ip,j,j,iq,n);
		      }
		  for(j=iq+1;j<n;j++)
		      {
		      ROTATE(a,ip,j,iq,j,n);
		      }
		  for(j=0;j<n;j++)
		      {
		      ROTATE(v,j,ip,j,iq,n);
		      }
		  ++(*nrot);
		  }
	      }
	  }
      for(ip=0;ip<n;++ip)
	  {
	  b[ip] += z[ip];
	  d[ip]=b[ip];
	  z[ip]=0.0;
	  }
      }

  /* Failed to converge!! What to do ??? */
  /* Well, let's at least free up memory and return without a murmur */

  delete [] b;
  delete [] z;
  return;
  
  }

qmVectorRef& 
qmEigenStructure::eigenSymmetric( const qmMatrixRef& symMat,
				  qmVectorRef& eValOut,
				  qmMatrixRef& eVecOut )
{
    int nrot;
    qmMatrix copy( symMat );
    // jacobi( copy, 3, eValOut, eVecOut, &nrot );
    jacobi( copy, symMat.getWidth(), eValOut, eVecOut, &nrot );
    return eValOut;
}
