// qmUnitQuaternionRef.c++

#include "QuickMath/qmUnitQuaternionRef.h++"
#include <math.h>
#include "QuickMath/qmMinMax.h++"
#include "QuickMath/qmVector.h++"

qmUnitQuaternionRef&
qmUnitQuaternionRef::set( real* rp )
{
    qmVectorRef::set( 4, rp ); 
    return *this;
}

qmMatrixRef&
qmUnitQuaternionRef::getRotationMatrix( qmMatrixRef& R ) const
{
  real q01 = array[0]*array[1];
  real q02 = array[0]*array[2];
  real q03 = array[0]*array[3];
  real q11 = array[1]*array[1];
  real q12 = array[1]*array[2];
  real q13 = array[1]*array[3];
  real q22 = array[2]*array[2];
  real q23 = array[2]*array[3];
  real q33 = array[3]*array[3];
  
  R(0,0) = 1.0 - 2 * (q22 + q33);
  R(0,1) =     - 2 * (q03 - q12);
  R(0,2) =       2 * (q02 + q13);
  R(1,0) =       2 * (q03 + q12);
  R(1,1) = 1.0 - 2 * (q11 + q33);
  R(1,2) =     - 2 * (q01 - q23);
  R(2,0) =     - 2 * (q02 - q13);
  R(2,1) =       2 * (q01 + q23);
  R(2,2) = 1.0 - 2 * (q11 + q22);

  return R;
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::setIdentity()
{
  array[0] = 1.;
  array[1] = 0.;
  array[2] = 0.;
  array[3] = 0.;
  return *this;
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::setMultiply( const qmUnitQuaternionRef& q1,
				  const qmUnitQuaternionRef& q2 )
{
  real qtemp[4];  // Allows for multiplication in place

  qtemp[0] = q1.array[0]*q2.array[0] - q1.array[1]*q2.array[1] - 
     q1.array[2]*q2.array[2] - q1.array[3]*q2.array[3];
  qtemp[1] = q1.array[1]*q2.array[0] + q1.array[0]*q2.array[1] - 
     q1.array[3]*q2.array[2] + q1.array[2]*q2.array[3];
  qtemp[2] = q1.array[2]*q2.array[0] + q1.array[3]*q2.array[1] + 
     q1.array[0]*q2.array[2] - q1.array[1]*q2.array[3];
  qtemp[3] = q1.array[3]*q2.array[0] - q1.array[2]*q2.array[1] + 
     q1.array[1]*q2.array[2] + q1.array[0]*q2.array[3];
  array[0] = qtemp[0];
  array[1] = qtemp[1];
  array[2] = qtemp[2];
  array[3] = qtemp[3];
  return *this;
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::setNormal()
{
  double mag = sqrt( (double)(array[0]*array[0] +
			      array[1]*array[1] +
			      array[2]*array[2] +
			      array[3]*array[3] ) );
  if( mag < .00000001 )
     return setIdentity();
  for( int i = 0; i < 4; i++ )
     array[i] /= mag;
  return *this;
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::setInvert( const qmUnitQuaternionRef& uq )
{
    array[0] = uq.array[0];
    array[1] = -uq.array[1];
    array[2] = -uq.array[2];
    array[3] = -uq.array[3];
    return *this;
}

static const real AngLim = .35; // 20 degrees

static
real
limitAngle( real angle )
{
  return limit( angle, (real)(-AngLim), (real)(AngLim) );
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::setAddAngles( const qmUnitQuaternionRef& uq,
				   const qmVectorRef& angles )
{
    qmVector qin = uq; // so you can add onto self
    qmVector dq(4);
    dq( 1 ) = limitAngle( angles.getValue(0) ) / 2.;
    dq( 2 ) = limitAngle( angles.getValue(1) ) / 2.;
    dq( 3 ) = limitAngle( angles.getValue(2) ) / 2.;
#ifdef QmWarnings
    // !!! warning: we really should warn when clipping happens
    if( dq(1) != angles.getValue(0)/2. || 
	dq(2) != angles.getValue(1)/2. || 
	dq(3) != angles.getValue(2)/2. )
       cerr << "qmUnitQuaternionRef::setAddAngles: CLIP: "
	    << angles.getValue(0) << ", " 
	    << angles.getValue(1) << ", " 
	    << angles.getValue(2) << endl;
#endif
    dq( 0 ) = 1. - dq(1)*dq(1) - dq(2)*dq(2) - dq(3)*dq(3);
    dq(0) = sqrt( fabs( (double) dq(0) ) );
    array[0] = qin(0)*dq(0) - qin(1)*dq(1) - qin(2)*dq(2) - qin(3)*dq(3);
    array[1] = qin(1)*dq(0) + qin(0)*dq(1) - qin(3)*dq(2) + qin(2)*dq(3);
    array[2] = qin(2)*dq(0) + qin(3)*dq(1) + qin(0)*dq(2) - qin(1)*dq(3);
    array[3] = qin(3)*dq(0) - qin(2)*dq(1) + qin(1)*dq(2) + qin(0)*dq(3);
    real mag = sqrt( array[0]*array[0] + array[1]*array[1] + 
		     array[2]*array[2] + array[3]*array[3] );
    for( int i = 0; i < 4; i++ )
       array[i] /= mag;
    return *this;
}

real
qmUnitQuaternionRef::getAngle()
{
    real cosineHalfAngle = this -> getValue( 0 );
    qmVectorRef imagPart = this -> subVector( 1, 3 );
    real sineHalfAngle = sqrt( imagPart.getInnerProduct( imagPart ) );
    real halfAngle = atan2( sineHalfAngle, cosineHalfAngle );
    return( 2*halfAngle );
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::setSubtract( const qmUnitQuaternionRef& q1,
				  const qmUnitQuaternionRef& q2 )
{
    real q2i[4];
    qmUnitQuaternionRef( q2i ).setInvert( q2 );
    return setMultiply( q1, qmUnitQuaternionRef( q2i ) );
}

qmUnitQuaternionRef&
qmUnitQuaternionRef::set( const qmMatrixRef& rin )
{
    static real m_data[9];
    qmVector qsquare( 4 );
    qmMatrixRef r( 3, 3, m_data);
    r.copy( rin );
    qsquare(0) = ( 1 + r(0,0) + r(1,1) + r(2,2) ) / 4.;
    qsquare(1) = ( 1 + r(0,0) - r(1,1) - r(2,2) ) / 4.;
    qsquare(2) = ( 1 - r(0,0) + r(1,1) - r(2,2) ) / 4.;
    qsquare(3) = ( 1 - r(0,0) - r(1,1) + r(2,2) ) / 4.;

    real maxsquare = 0.;
    int maxi = 0;
    for( int i = 0; i < 4; i++ )
       if( qsquare(i) > maxsquare )
	  {
	      maxsquare = qsquare(i);
	      maxi = i;
	  }
    real maxval = sqrt( maxsquare );
    array[maxi] = maxval;

    switch( maxi )
       {
	  case 0:
	   array[1] = (r(2,1)-r(1,2))/(4.*maxval);
	   array[2] = (r(0,2)-r(2,0))/(4.*maxval);
	   array[3] = (r(1,0)-r(0,1))/(4.*maxval);
	   break;
	  case 1:
	   array[0] = (r(2,1)-r(1,2))/(4.*maxval);
	   array[2] = (r(1,0)+r(0,1))/(4.*maxval);
	   array[3] = (r(0,2)+r(2,0))/(4.*maxval);
	   break;
	  case 2:
	   array[0] = (r(0,2)-r(2,0))/(4.*maxval);
	   array[1] = (r(1,0)+r(0,1))/(4.*maxval);
	   array[3] = (r(2,1)+r(1,2))/(4.*maxval);
	   break;
	  case 3:
	   array[0] = (r(1,0)-r(0,1))/(4.*maxval);
	   array[1] = (r(0,2)+r(2,0))/(4.*maxval);
	   array[2] = (r(2,1)+r(1,2))/(4.*maxval);
	   break;
       }
    if( array[0] < 0 )
       setNegateEq();
    return *this;
}
