// qmKalmanFilter.c++
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
// at the Media Laboratory, MIT, Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.
// 

#include "qmKalmanFilter.h++"

const float qmKalmanFilter::veryBig = 1000000000.;

qmKalmanFilter::qmKalmanFilter( int nstates,
				int nmmts,
				qmVectorRef& initialState,
				qmMatrixRef& initialCovariance,
				qmMatrixRef& mmtModel,
				qmMatrixRef& mmtNoise,
				qmMatrixRef& dynModel,
				qmMatrixRef& dynNoise )
  : n( nstates ),
    p( nmmts ),
    t( 0 ),
    xP( nstates ),		// init
    xE( nstates ),
    dX( nstates ),
    yP( nmmts ),
    dY( nmmts ),
    Pp( nstates, nstates ),	// init
    Pe( nstates, nstates ),
    K( nstates, nmmts ),
    I( nstates, nstates ),	// init
    tempNxP( nstates, nmmts ),
    tempNxN( nstates, nstates ),
    tempPxP( nmmts, nmmts ),
    H( nmmts, nstates ),	// init
    Phi( nstates, nstates ),	// init
    Q( nstates, nstates ),	// init
    R( nmmts, nmmts )		// init
{
    xP = initialState;
    Pp = initialCovariance;
    H = mmtModel;
    R = mmtNoise;
    Phi = dynModel;
    Q = dynNoise;
    I.setIdentity();
}

inline
qmMatrixRef&
_makeStateTransition( qmMatrixRef& phi, int n, float dt )
{
    // assume phi is properly zeroed, and 1 on diagonal
    
    float halfDtSquared = .5 * dt * dt;
    
    for( int i = 0; i < n; i++ )
       {
	   phi( i, i+n ) = dt;
	   phi( i, i+2*n ) = halfDtSquared;
	   phi( i+n, i+2*n ) = dt;
       }
    return phi;
}

qmKalmanFilter::qmKalmanFilter( int ns, float rv, float qv, float dt )
  : n( 3 * ns ),
    p( ns ),
    t( 0 ),
    xP( n ),	// init
    xE( n ),
    dX( n ),
    yP( p ),
    dY( p ),
    Pp( n, n ),	// init
    Pe( n, n ),
    K( n, p ),
    I( n, n ),	// init
    tempNxP( n, p ),
    tempNxN( n, n ),
    tempPxP( p, p ),
    H( p, n ),	// init
    Phi( n, n ),// init
    Q( n, n ),	// init
    R( p, p )	// init
{
    xP.setZero();
    Pp.setDiagonal( veryBig );
    Pe.setDiagonal( veryBig );
    H.setZero();
    for( int i = 0; i < ns; i++ )
       H( i, i ) = 1.;
    R.setDiagonal( rv*rv );
    Phi.setIdentity();
    _makeStateTransition( Phi, ns, dt );
    Q.setZero();
    for( i = 0; i < ns; i++ )
       Q( i+2*ns, i+2*ns ) = qv*qv;
    I.setIdentity();
}

qmKalmanFilter&
qmKalmanFilter::update( qmVectorRef& y )
{
    // ---- state prediction
    xP.setMultiply( Phi, xE );
    // ---- error covariance for prediction
    Pp.setMultiplyT( tempNxN.setMultiply( Phi, Pe ), Phi ).setAddEq( Q );
    // ---- Kalman gain
    tempNxP.setMultiplyT( Pp, H );
    tempPxP.setMultiply( H, tempNxP );
    tempPxP.setAddEq( R );
    tempPxP.setInvertEq();
    K.setMultiply( tempNxP, tempPxP );
    // ---- innovation
    yP.setMultiply( H, xP );
    dY.setSubtract( y, yP );
    // ---- state estimate
    dX.setMultiply( K, dY );
    xE.setAdd( xP, dX );
    // ---- error covariance for estimate
    Pe.setMultiply
       ( tempNxN.setMultiply( K, H ).setSubtract( I, tempNxN ), Pp );
    //  does this work?
    return *this;
}
